/***************************************************************************
                          MeshFunction_impl.h  -  description
                             -------------------
    begin                : Nov 8, 2015
    copyright            : (C) 2015 by oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#include <TNL/Assert.h>
#include <TNL/DevicePointer.h>
#include <TNL/Functions/MeshFunction.h>
#include <TNL/Functions/MeshFunctionEvaluator.h>
#include <TNL/Functions/MeshFunctionNormGetter.h>
#include <TNL/Functions/MeshFunctionGnuplotWriter.h>
#include <TNL/Functions/MeshFunctionVTKWriter.h>

#pragma once

namespace TNL {
namespace Functions {   

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
MeshFunction()
{
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
MeshFunction( const MeshPointer& meshPointer )
: meshPointer( meshPointer )
{
   this->data.setSize( getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
MeshFunction( const ThisType& meshFunction )
: meshPointer( meshFunction.meshPointer )
{
   this->data.bind( meshFunction.getData() );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Vector >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
MeshFunction( const MeshPointer& meshPointer,
              Vector& data,
              const IndexType& offset )
: meshPointer( meshPointer )
{
   this->data.bind( data, offset, getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );   
}


template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Vector >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
MeshFunction( const MeshPointer& meshPointer,
              SharedPointer< Vector >& data,
              const IndexType& offset )
: meshPointer( meshPointer )
{
   this->data.bind( *data, offset, getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );   
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
String
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getType()
{
   return String( "Functions::MeshFunction< " ) +
                     Mesh::getType() + ", " +
                     String( MeshEntityDimension ) + ", " +
                    TNL::getType< Real >() +
                     " >";
};

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
String
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getTypeVirtual() const
{
   return this->getType();
};

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
String
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getSerializationType()
{
   return String( "Functions::MeshFunction< " ) +
                     Mesh::getSerializationType() + ", " +
                     String( MeshEntityDimension ) + ", " +
                    TNL::getType< Real >() +
                     " >";
};

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
String
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getSerializationTypeVirtual() const
{
   return this->getSerializationType();
};

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
configSetup( Config::ConfigDescription& config,
             const String& prefix )
{
   config.addEntry< String >( prefix + "file", "Dataset for the mesh function." );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
setup( const MeshPointer& meshPointer,
       const Config::ParameterContainer& parameters,
       const String& prefix )
{
   this->setMesh( meshPointer );
   if( parameters.checkParameter( prefix + "file" ) )
   {
      String fileName = parameters.getParameter< String >( prefix + "file" );
      if( ! this->load( fileName ) )
         return false;
   }
   else
   {
      std::cerr << "Missing parameter " << prefix << "file." << std::endl;
      throw(0);
      return false;
   }
   return true;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
bind( ThisType& meshFunction )
{
   this->meshPointer = meshFunction.getMeshPointer();
   this->data.bind( meshFunction.getData() );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Vector >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
bind( const MeshPointer& meshPointer,
      const Vector& data,
      const IndexType& offset )
{
   this->meshPointer = meshPointer;
   this->data.bind( data, offset, getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() * this->getRangeDimension() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() * this->getRangeDimension(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );   
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Vector >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
bind( const MeshPointer& meshPointer,
      const SharedPointer< Vector >& data,
      const IndexType& offset )
{
   this->meshPointer = meshPointer;
   this->data.bind( *data, offset, getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() * this->getRangeDimension() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() * this->getRangeDimension(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );   
}


template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
setMesh( const MeshPointer& meshPointer )
{
   this->meshPointer = meshPointer;
   this->data.setSize( getMesh().template getEntitiesCount< typename Mesh::template EntityType< MeshEntityDimension > >() * this->getRangeDimension() );
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() * this->getRangeDimension(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );   
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
 template< typename Device >
__cuda_callable__
const typename MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::MeshType& 
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getMesh() const
{
   return this->meshPointer.template getData< Device >();
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
const typename MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::MeshPointer&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getMeshPointer() const
{
   return this->meshPointer;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
typename MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::IndexType
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getDofs( const MeshPointer& meshPointer )
{
   return meshPointer->template getEntitiesCount< getEntitiesDimension() >();
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
__cuda_callable__
const typename MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::VectorType& 
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getData() const
{
   return this->data;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
__cuda_callable__
typename MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::VectorType& 
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getData()
{
   return this->data;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
refresh( const RealType& time ) const
{
   return true;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
deepRefresh( const RealType& time ) const
{
   return true;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename EntityType >
typename Functions::MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::RealType
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getValue( const EntityType& meshEntity, const int& component ) const
{
   static_assert( EntityType::getEntityDimension() == MeshEntityDimension, "Calling with wrong EntityType -- entity dimensions do not match." );
   TNL_ASSERT( component >= 0 && component < this->getRangeDimension(),
               std::cerr << "Wrong component number: " << component << ". The function dimension is " << this->getRangeDimension() << "." << std::endl; );
   const IndexType entities = this->getMesh().template getEntitiesCount< EntityType >();
   return this->data.getElement( component * entities + meshEntity.getIndex() );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename EntityType >
void
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
setValue( const EntityType& meshEntity,
          const RealType& value,
          const int& component )
{
   static_assert( EntityType::getEntityDimension() == MeshEntityDimension, "Calling with wrong EntityType -- entity dimensions do not match." );
   TNL_ASSERT( component >= 0 && component < this->getRangeDimension(),
               std::cerr << "Wrong component number: " << component << ". The function dimension is " << this->getRangeDimension() << "." << std::endl; );
   const IndexType entities = this->getMesh().template getEntitiesCount< EntityType >();
   this->data.setElement( component * entities + meshEntity.getIndex(), value );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename EntityType >
__cuda_callable__
typename Functions::MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::RealType&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator()( const EntityType& meshEntity,
            const RealType& time,
            const int& component )
{
   static_assert( EntityType::getEntityDimension() == MeshEntityDimension, "Calling with wrong EntityType -- entity dimensions do not match." );
   TNL_ASSERT( component >= 0 && component < this->getRangeDimension(),
               std::cerr << "Wrong component number: " << component << ". The function dimension is " << this->getRangeDimension() << "." << std::endl; );
   const IndexType entities = this->template getMesh< DeviceType >().template getEntitiesCount< EntityType >();
   return this->data[ component * entities + meshEntity.getIndex() ];
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename EntityType >
__cuda_callable__
const typename Functions::MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::RealType&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator()( const EntityType& meshEntity,
            const RealType& time,
            const int& component ) const
{
   static_assert( EntityType::getEntityDimension() == MeshEntityDimension, "Calling with wrong EntityType -- entity dimensions do not match." );
   TNL_ASSERT( component >= 0 && component < this->getRangeDimension(),
               std::cerr << "Wrong component number: " << component << ". The function dimension is " << this->getRangeDimension() << "." << std::endl; );
   const IndexType entities = this->template getMesh< DeviceType >().template getEntitiesCount< EntityType >();
   return this->data[ component * entities + meshEntity.getIndex() ];
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
__cuda_callable__
typename Functions::MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::RealType&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator[]( const IndexType& meshEntityIndex )
{
   return this->data[ meshEntityIndex ];
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
__cuda_callable__
const typename Functions::MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::RealType&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator[]( const IndexType& meshEntityIndex ) const
{
   return this->data[ meshEntityIndex ];
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Function >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator = ( const Function& f )
{
   static_assert( ThisType::getRangeDimension() == Function::getRangeDimension(), "Only functions of the same dimension can be assigned." );
   DevicePointer< ThisType > thisDevicePtr( *this );
   DevicePointer< typename std::add_const< Function >::type > fDevicePtr( f );
   MeshFunctionEvaluator< ThisType, Function > evaluator;
   evaluator.evaluate( thisDevicePtr, fDevicePtr );
   return *this;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Function >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator += ( const Function& f )
{
   static_assert( ThisType::getRangeDimension() == Function::getRangeDimension(), "Only functions of the same dimension can be added." );
   DevicePointer< ThisType > thisDevicePtr( *this );
   DevicePointer< typename std::add_const< Function >::type > fDevicePtr( f );
   MeshFunctionEvaluator< ThisType, Function > evaluator;
   evaluator.evaluate( thisDevicePtr, fDevicePtr, 1.0, 1.0 );
   return *this;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
   template< typename Function >
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >&
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
operator -= ( const Function& f )
{
   static_assert( ThisType::getRangeDimension() == Function::getRangeDimension(), "Only functions of the same dimension can be subtracted." );
   DevicePointer< ThisType > thisDevicePtr( *this );
   DevicePointer< typename std::add_const< Function >::type > fDevicePtr( f );
   MeshFunctionEvaluator< ThisType, Function > evaluator;
   evaluator.evaluate( thisDevicePtr, fDevicePtr, 1.0, -1.0 );
   return *this;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
Real
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getLpNorm( const RealType& p, const int& component ) const
{
   // TODO: return norm of the component
   return MeshFunctionNormGetter< ThisType >::getNorm( *this, p );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
Real
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
getMaxNorm( const int& component ) const
{
   // TODO: return norm of the component
   return this->data.absMax();
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
save( File& file ) const
{
   TNL_ASSERT( this->data.getSize() == this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() * this->getRangeDimension(), 
               std::cerr << "this->data.getSize() = " << this->data.getSize() << std::endl
                         << "this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() = " << this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >() );
   if( ! Object::save( file ) )
      return false;
   return this->data.save( file );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
load( File& file )
{
   if( ! Object::load( file ) )
      return false;
   if( ! this->data.load( file ) )
      return false;
   const IndexType meshSize = this->getMesh().template getEntitiesCount< typename MeshType::template EntityType< MeshEntityDimension > >();
   if( this->data.getSize() != meshSize * this->getRangeDimension() )
   {      
      std::cerr << "Size of the data loaded to the mesh function (" << this->data.getSize() << ") does not fit with the mesh size (" << meshSize << ") times range dimensions (" << this->getRangeDimension() << ")." << std::endl;
      return false;
   }
   return true;
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
boundLoad( File& file )
{
   if( ! Object::load( file ) )
      return false;
   return this->data.boundLoad( file );
}

template< typename Mesh,
          int MeshEntityDimension,
          typename Real,
          int MeshFunctionDimension >
bool
MeshFunction< Mesh, MeshEntityDimension, Real, MeshFunctionDimension >::
write( const String& fileName,
       const String& format,
       const double& scale ) const
{
   std::fstream file;
   file.open( fileName.getString(), std::ios::out );
   if( ! file )
   {
      std::cerr << "Unable to open a file " << fileName << "." << std::endl;
      return false;
   }
   if( format == "vtk" )
      return MeshFunctionVTKWriter< ThisType >::write( *this, file, scale );
   else if( format == "gnuplot" )
      return MeshFunctionGnuplotWriter< ThisType >::write( *this, file, scale );
   else {
      std::cerr << "Unknown output format: " << format << std::endl;
      return false;
   }
   return true;
}
 
} // namespace Functions
} // namespace TNL

