/***************************************************************************
                          Operator.h  -  description
                             -------------------
    begin                : Aug 16, 2016
    copyright            : (C) 2016 by Tomas Oberhuber
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <type_traits>

#include <TNL/Devices/Cuda.h>
#include <TNL/function_traits.h>

namespace TNL {
namespace Operators {   

// implementation for 1x1 operators
// (which either specify getPreimageComponents() == 1 && getImageComponents() == 1,
// or don't implement these methods at all)
template< typename Operator, typename Enable = void >
class OperatorAdapter
{
   public:
      using RealType = typename Operator::RealType;
      using IndexType = typename Operator::IndexType;
      using MeshType = typename Operator::MeshType;

      template< typename EntityType,
                typename MeshFunction >
      __cuda_callable__
      const RealType operator()( const Operator& op,
                                 const MeshFunction& u,
                                 const EntityType& entity,
                                 const RealType& time,
                                 const int& component ) const
      {
         TNL_ASSERT( component == 0, );
         return op( u, entity, time );
      }

      template< typename EntityType >
      __cuda_callable__
      IndexType getLinearSystemRowLength( const Operator& op,
                                          const MeshType& mesh,
                                          const IndexType& index,
                                          const EntityType& entity,
                                          const int& component ) const
      {
         return op.getLinearSystemRowLength( mesh, index, entity );
      }

      template< typename PreimageFunction,
                typename MeshEntity,
                typename Matrix,
                typename Vector >
      __cuda_callable__
      void setMatrixElements( const Operator& op,
                              const PreimageFunction& u,
                              const MeshEntity& entity,
                              const RealType& time,
                              const RealType& tau,
                              const int& component,
                              Matrix& matrix,
                              Vector& b ) const
      {
         TNL_ASSERT( component == 0, );
         op.setMatrixElements( u, entity, time, tau, matrix, b );
      }
};

template< typename Operator >
//class OperatorAdapter< Operator, typename std::enable_if< Operator::getPreimageComponents() >= 2 || Operator::getImageComponents() >= 2 >::type >
// error: decltype cannot resolve address of overloaded function
//class OperatorAdapter< Operator, typename std::enable_if< function_traits< Operator >::arity == 4 >::type >
class OperatorAdapter< Operator, typename std::enable_if< function_traits< decltype(&Operator::getLinearSystemRowLength) >::arity == 5 >::type >
{
   public:
      using RealType = typename Operator::RealType;
      using IndexType = typename Operator::IndexType;
      using MeshType = typename Operator::MeshType;

      template< typename EntityType,
                typename MeshFunction >
      __cuda_callable__ inline
      const RealType operator()( const Operator& op,
                                 const MeshFunction& u,
                                 const EntityType& entity,
                                 const RealType& time,
                                 const int& component ) const
      {
         return op( u, entity, time, component );
      }

      template< typename EntityType >
      __cuda_callable__
      IndexType getLinearSystemRowLength( const Operator& op,
                                          const MeshType& mesh,
                                          const IndexType& index,
                                          const EntityType& entity,
                                          const int& component ) const
      {
         return op.getLinearSystemRowLength( mesh, index, entity, component );
      }

      template< typename PreimageFunction,
                typename MeshEntity,
                typename Matrix,
                typename Vector >
      __cuda_callable__ inline
      void setMatrixElements( const Operator& op,
                              const PreimageFunction& u,
                              const MeshEntity& entity,
                              const RealType& time,
                              const RealType& tau,
                              const int& component,
                              Matrix& matrix,
                              Vector& b ) const
      {
         op.setMatrixElements( u, entity, time, tau, component, matrix, b );
      }
};

} // namespace Operators
} // namespace TNL

