/***************************************************************************
                          OperatorEvaluator.h  -  description
                             -------------------
    begin                : Dec 21, 2016
    copyright            : (C) 2016 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <TNL/SharedPointer.h>

namespace TNL {
namespace Operators {   

// TODO: figure out how to store arbitrary parameters (specified as variadic template parameters) in the UserData
template< typename InteriorOperator,
          typename BoundaryOperator,
          typename Matrix >
class OperatorEvaluatorTraverserUserData
{
   static_assert( InteriorOperator::getImageComponents() == BoundaryOperator::getImageComponents(),
                  "The InteriorOperator and BoundaryOperator must have the same number of image components." );

   public:
      const InteriorOperator* interiorOperator = NULL;

      const BoundaryOperator* boundaryOperator = NULL;

      Matrix* matrix = NULL;

      void setUserData( const InteriorOperator* interiorOperator,
                        const BoundaryOperator* boundaryOperator,
                        Matrix* matrix )
      {
         this->interiorOperator = interiorOperator;
         this->boundaryOperator = boundaryOperator;
         this->matrix = matrix;
      }
};


template< typename Mesh,
          typename InteriorOperator,
          typename BoundaryOperator,
          typename Matrix >
class OperatorEvaluator
{
public:
   using MeshType = Mesh;
   using DeviceType = typename MeshType::DeviceType;
   using MatrixType = Matrix;
   using TraverserUserData = OperatorEvaluatorTraverserUserData< InteriorOperator,
                                                                 BoundaryOperator,
                                                                 MatrixType >;
   using MeshPointer = SharedPointer< MeshType >;
   using InteriorOperatorPointer = SharedPointer< InteriorOperator, DeviceType >;
   using BoundaryOperatorPointer = SharedPointer< BoundaryOperator, DeviceType >;
   using MatrixPointer = SharedPointer< Matrix, DeviceType >;
   
      
   template< typename EntityType >
   void evaluate( const MeshPointer& meshPointer,
                  const InteriorOperatorPointer& interiorOperatorPointer,
                  const BoundaryOperatorPointer& boundaryOperatorPointer,
                  MatrixPointer& matrixPointer );

 
   class TraverserInteriorEntitiesProcessor
   {
      public:

         template< typename EntityType >
         __cuda_callable__
         static void processEntity( const MeshType& mesh,
                                    TraverserUserData& userData,
                                    const EntityType& entity,
                                    const int& component )
         {
            ( *userData.interiorOperator ).setMatrixElements
               ( mesh,
                 entity,
                 component,
                 ( *userData.matrix ) );
         }
   };

   class TraverserBoundaryEntitiesProcessor
   {
      public:
 
         template< typename EntityType >
         __cuda_callable__
         static void processEntity( const MeshType& mesh,
                                    TraverserUserData& userData,
                                    const EntityType& entity,
                                    const int& component )
         {
            ( *userData.boundaryOperator ).setMatrixElements
               ( mesh,
                 entity,
                 component,
                 ( *userData.matrix ) );
         }
   };

protected:
   SharedPointer< TraverserUserData, DeviceType > userDataPointer;
};

} // namespace Operators
} // namespace TNL

#include <TNL/Operators/OperatorEvaluator_impl.h>
