/***************************************************************************
                          OperatorEvaluator_impl.h  -  description
                             -------------------
    begin                : Dec 21, 2016
    copyright            : (C) 2016 by Tomas Oberhuber et al.
    email                : tomas.oberhuber@fjfi.cvut.cz
 ***************************************************************************/

/* See Copyright Notice in tnl/Copyright */

#pragma once

#include <type_traits>
#include <TNL/Meshes/Traverser.h>
#include <TNL/Operators/OperatorEvaluator.h>

namespace TNL {
namespace Operators {   

template< typename Mesh,
          typename InteriorOperator,
          typename BoundaryOperator,
          typename Matrix >
   template< typename EntityType >
void
OperatorEvaluator< Mesh, InteriorOperator, BoundaryOperator, Matrix >::
evaluate( const MeshPointer& meshPointer,
          const InteriorOperatorPointer& interiorOperatorPointer,
          const BoundaryOperatorPointer& boundaryOperatorPointer,
          MatrixPointer& matrixPointer )
{
   const auto maxRowLength = matrixPointer.template getData< Devices::Host >().getMaxRowLength();
   TNL_ASSERT( maxRowLength > 0, );

   this->userDataPointer->setUserData(
         &interiorOperatorPointer.template getData< DeviceType >(),
         &boundaryOperatorPointer.template getData< DeviceType >(),
         &matrixPointer.template modifyData< DeviceType >() );
   Meshes::Traverser< MeshType, EntityType, InteriorOperator::getImageComponents() > meshTraverser;
   meshTraverser.template processBoundaryEntities< TraverserUserData,
                                                   TraverserBoundaryEntitiesProcessor >
                                                 ( meshPointer,
                                                   userDataPointer );
   meshTraverser.template processInteriorEntities< TraverserUserData,
                                                   TraverserInteriorEntitiesProcessor >
                                                 ( meshPointer,
                                                   userDataPointer );
}

} // namespace Operators
} // namespace TNL
